cgc -profile ps_2_0 -entry p_AdjustVelocity -o AdjustVelocity.ps_2_0.txt AdjustVelocity.fx
cgc -profile ps_3_0 -entry p_AdjustVelocity -o AdjustVelocity.ps_3_0.txt AdjustVelocity.fx
cgc -profile arbfp1 -entry p_AdjustVelocity -o AdjustVelocity.arbfp1.txt AdjustVelocity.fx
cgc -profile ps_2_0 -entry p_BoundaryMixed -o BoundaryMixed.ps_2_0.txt BoundaryMixed.fx
cgc -profile ps_3_0 -entry p_BoundaryMixed -o BoundaryMixed.ps_3_0.txt BoundaryMixed.fx
cgc -profile arbfp1 -entry p_BoundaryMixed -o BoundaryMixed.arbfp1.txt BoundaryMixed.fx
cgc -profile ps_2_0 -entry p_CopyState -o CopyState.ps_2_0.txt CopyState.fx
cgc -profile ps_3_0 -entry p_CopyState -o CopyState.ps_3_0.txt CopyState.fx
cgc -profile arbfp1 -entry p_CopyState -o CopyState.arbfp1.txt CopyState.fx
cgc -profile ps_2_0 -entry p_Divergence -o Divergence.ps_2_0.txt Divergence.fx
cgc -profile ps_3_0 -entry p_Divergence -o Divergence.ps_3_0.txt Divergence.fx
cgc -profile arbfp1 -entry p_Divergence -o Divergence.arbfp1.txt Divergence.fx
cgc -profile ps_2_0 -entry p_DrawDensity -o DrawDensity.ps_2_0.txt DrawDensity.fx
cgc -profile ps_3_0 -entry p_DrawDensity -o DrawDensity.ps_3_0.txt DrawDensity.fx
cgc -profile arbfp1 -entry p_DrawDensity -o DrawDensity.arbfp1.txt DrawDensity.fx
cgc -profile ps_2_0 -entry p_FluidUpdate -o FluidUpdate.ps_2_0.txt FluidUpdate.fx
cgc -profile ps_3_0 -entry p_FluidUpdate -o FluidUpdate.ps_3_0.txt FluidUpdate.fx
cgc -profile arbfp1 -entry p_FluidUpdate -o FluidUpdate.arbfp1.txt FluidUpdate.fx
cgc -profile ps_2_0 -entry p_PoissonSolver -o PoissonSolver.ps_2_0.txt PoissonSolver.fx
cgc -profile ps_3_0 -entry p_PoissonSolver -o PoissonSolver.ps_3_0.txt PoissonSolver.fx
cgc -profile arbfp1 -entry p_PoissonSolver -o PoissonSolver.arbfp1.txt PoissonSolver.fx
