// Geometric Tools, LLC
// Copyright (c) 1998-2014
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt
// http://www.geometrictools.com/License/Boost/LICENSE_1_0.txt
//
// File Version: 5.3.0 (2010/09/07)

#ifndef SAVEBMP32_H
#define SAVEBMP32_H

#include "Image2.h"
#include "PixelBGRA8.h"

bool SaveBMP32 (const std::string& name, const Image2<PixelBGRA8>& image);

#endif
