// Geometric Tools, LLC
// Copyright (c) 1998-2014
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt
// http://www.geometrictools.com/License/Boost/LICENSE_1_0.txt
//
// File Version: 5.6.0 (2011/07/23)

#ifndef WM5GLUTRENDERERDATA_H
#define WM5GLUTRENDERERDATA_H

#include "Wm5OpenGLRendererLIB.h"
#include "Wm5OpenGLRendererData.h"

namespace Wm5
{

class GlutRendererData : public RendererData
{
public:
    int mWindowID;
};

}

#endif
