// Geometric Tools, LLC
// Copyright (c) 1998-2014
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt
// http://www.geometrictools.com/License/Boost/LICENSE_1_0.txt
//
// File Version: 5.0.0 (2010/01/01)

#ifndef RATIONAL_H
#define RATIONAL_H

#include "Wm5RVector2.h"
using namespace Wm5;

typedef Rational<4*sizeof(float)> RScalar;
typedef RVector2<4*sizeof(float)> RPoint2;

#endif
