// Geometric Tools, LLC
// Copyright (c) 1998-2014
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt
// http://www.geometrictools.com/License/Boost/LICENSE_1_0.txt
//
// File Version: 5.0.0 (2010/01/01)

#include "Wm5GraphicsPCH.h"
#include "Wm5VertexBuffer.h"
#include "Wm5Renderer.h"
using namespace Wm5;

WM5_IMPLEMENT_RTTI(Wm5, Buffer, VertexBuffer);
WM5_IMPLEMENT_STREAM(VertexBuffer);
WM5_IMPLEMENT_FACTORY(VertexBuffer);
WM5_IMPLEMENT_DEFAULT_NAMES(Buffer, VertexBuffer);
WM5_IMPLEMENT_DEFAULT_STREAM(Buffer, VertexBuffer);

//----------------------------------------------------------------------------
VertexBuffer::VertexBuffer (int numVertices, int vertexSize, Usage usage)
    :
    Buffer(numVertices, vertexSize, usage)
{
}
//----------------------------------------------------------------------------
VertexBuffer::~VertexBuffer ()
{
    Renderer::UnbindAll(this);
}
//----------------------------------------------------------------------------
