// Geometric Tools, LLC
// Copyright (c) 1998-2014
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt
// http://www.geometrictools.com/License/Boost/LICENSE_1_0.txt
//
// File Version: 5.0.1 (2010/10/01)

#ifndef WM5IMAGECONVERT_H
#define WM5IMAGECONVERT_H

#include "Wm5ImagicsLIB.h"

namespace Wm5
{

WM5_IMAGICS_ITEM void ImageConvert (int quantity, int srcRTTI, void* srcData,
    int trgRTTI, void* trgData);

}

#endif
