// Geometric Tools, LLC
// Copyright (c) 1998-2014
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt
// http://www.geometrictools.com/License/Boost/LICENSE_1_0.txt
//
// File Version: 5.0.0 (2010/01/01)

#ifndef RIGIDBALL_H
#define RIGIDBALL_H

#include "Wm5RigidBody.h"
#include "Wm5TriMesh.h"
using namespace Wm5;

class RigidBall : public RigidBodyf
{
public:
    RigidBall (float radius = 1.0f);

    TriMeshPtr& Mesh ();
    float GetRadius () const;

    bool Moved;

private:
    TriMeshPtr mMesh;
    float mRadius;
};

#endif
