// Geometric Tools, LLC
// Copyright (c) 1998-2014
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt
// http://www.geometrictools.com/License/Boost/LICENSE_1_0.txt
//
// File Version: 5.0.0 (2010/01/01)

#include "Wm5ImagicsPCH.h"
#include "Wm5Element.h"
using namespace Wm5;

WM5_IMPL_ELEMENT(char, 0);
WM5_IMPL_ELEMENT(uchar, 1);
WM5_IMPL_ELEMENT(short, 2);
WM5_IMPL_ELEMENT(ushort, 3);
WM5_IMPL_ELEMENT(int, 4);
WM5_IMPL_ELEMENT(uint, 5);
WM5_IMPL_ELEMENT(long, 6);
WM5_IMPL_ELEMENT(ulong, 7);
WM5_IMPL_ELEMENT(float, 8);
WM5_IMPL_ELEMENT(double, 9);
WM5_IMPL_ELEMENT(rgb5, 10);
WM5_IMPL_ELEMENT(rgb8, 11);
