// Geometric Tools, LLC
// Copyright (c) 1998-2014
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt
// http://www.geometrictools.com/License/Boost/LICENSE_1_0.txt
//
// File Version: 5.5.0 (2010/10/23)

#ifndef MERGEBOX_H
#define MERGEBOX_H

class MergeBox
{
public:
    MergeBox (int stride)
        :
        mXStride(stride),
        mYStride(stride),
        mZStride(stride),
        mValid(true)
    {
    }

    int mXStride, mYStride, mZStride;
    bool mValid;
};

#endif
