// Geometric Tools, LLC
// Copyright (c) 1998-2014
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt
// http://www.geometrictools.com/License/Boost/LICENSE_1_0.txt
//
// File Version: 5.0.0 (2010/01/01)

#ifndef EXTRACTRIDGES_H
#define EXTRACTRIDGES_H

#include "Wm5ConsoleApplication.h"
using namespace Wm5;

class ExtractRidges : public ConsoleApplication
{
    WM5_DECLARE_INITIALIZE;
    WM5_DECLARE_TERMINATE;

public:
    ExtractRidges ();

    virtual int Main (int numArguments, char** arguments);

protected:
};

WM5_REGISTER_INITIALIZE(ExtractRidges);
WM5_REGISTER_TERMINATE(ExtractRidges);

#endif
